use strict;
use Math::Round qw(:all);
use List::Util qw( min max );
# ==============TO DELETE ========================
# TMP
#use Data::Dumper qw(Dumper);
#my $fl = ">TMP.txt";
#open (OUT,$fl) || die "Cannot open '$fl': $!";
# ================================================

#my $fileName = "OriginalModel.txt";
my $fileName = "AlberModelHet.txt";

chomp( my @L2 = `cat $fileName | awk '{if(\$2=="2L")print \$0}'` );
chomp( my @L3 = `cat $fileName | awk '{if(\$2=="3L")print \$0}'` );
chomp( my @R2 = `cat $fileName | awk '{if(\$2=="2R")print \$0}'` );
chomp( my @R3 = `cat $fileName | awk '{if(\$2=="3R")print \$0}'` );
chomp( my @X  = `cat $fileName | awk 'NR>1{if(\$2=="X")print \$0}'` );
chomp( my @C4 = `cat $fileName | awk 'NR>1{if(\$2=="4")print \$0}'` );

# use this block for the Pickersgil data (lam_positions.txt)
#chomp( my @L2lam = `cat lam_positions.txt | awk '{if(\$1=="L2")print \$2}'` );
#chomp( my @L3lam = `cat lam_positions.txt | awk '{if(\$1=="L3")print \$2}'` );
#chomp( my @R2lam = `cat lam_positions.txt | awk '{if(\$1=="R2")print \$2}'` );
#chomp( my @R3lam = `cat lam_positions.txt | awk '{if(\$1=="R3")print \$2}'` );
#chomp( my @Xlam  = `cat lam_positions.txt | awk '{if(\$1=="X")print \$2}'` );
#my @C4lam = (-1);

# use this block for the Bemmel data (Bemmel_lad.txt)
chomp( my @L2lam = `cat Bemmel_lad.txt | awk '{if(\$1=="2L")print (\$4+\$5)/2}'` );
chomp( my @R2lam = `cat Bemmel_lad.txt | awk '{if(\$1=="2R")print (\$4+\$5)/2}'` );
chomp( my @L3lam = `cat Bemmel_lad.txt | awk '{if(\$1=="3L")print (\$4+\$5)/2}'` );
chomp( my @R3lam = `cat Bemmel_lad.txt | awk '{if(\$1=="3R")print (\$4+\$5)/2}'` );
chomp( my @Xlam = `cat Bemmel_lad.txt | awk '{if(\$1=="X")print (\$4+\$5)/2}'` );
chomp( my @C4lam = `cat Bemmel_lad.txt | awk '{if(\$1=="4")print (\$4+\$5)/2}'` );

# ===============================================================================================
# Some parameters (don't change any except $roundRadiusDigit, $nucleus_radius, $nucleolus_radius)
# -----------------------------------------------------------------------------------------------
my $basePerNucleosome = 200; # bound + linker DNA in base pairs
my $NucleosomeDaltons = 132500;	# Nick has 100000
my $basePairDaltons = 660;	# Nick has 600
my $daltonToGram = 1.66054e-24;
my $roundRadiusDigit = 10;
# ==========================================================================
# Nick's nucleus model parameters, in micrometers
#  -----------------------------------------------
#my $nucleus_radius = 2.25;
#my $nucleolus_radius = .2;
#my $corcoeff = 1.410784; # to get 0.157 volume ratio with Alber's TAD beads
# ==========================================================================
# Alber's nucleus model parametes, in micrometers
# ------------------------------------------------
my $nucleus_radius = 2.00 + 0.04;
my $nucleolus_radius = .333;
my $corcoeff = 1.254031; # to get 0.157 volume ratio with Alber's TAD beads
# ==========================================================================
# Simulation parameters
# ---------------------
my $autoGeneratePoints = 1; #Should be 1.  If = 0, (matlab -nodesktop -nosplash < make_structures.m) is involved
my $output_warmup = 1;      #If 1, the trj writing command writevcf \$traj_file is included into the tcl code
my $envelope_attachments_on = 1; # If 1 - attachments are ON, if 0 - OFF
my $tads_on = 1;  # NEW, turns on specific TAD-TAD interactions excluding that pairs from non-specific interactions.
my %tad_tad = (); # hash table with the values from Sexton's 2012 table (need to be done)
my $box_length = 100;
my $output_frames = 400000; # for test purposes. Sould be increased to ~100,000
my $steps_per_frame = 1000;  # integration steps per one frame
my $colorByAttach = 1;

# ==========================================================================
# Our interaction parameters
# ---------------------------
my $eps_Null = 1.5; #label_type = 0
my $eps_Act  = 0.1; #label_type = 1
my $eps_PcG  = 1.5; #label_type = 2
my $eps_HP1  = 1.5; #label_type = 3
my $eps_gen  = 0.5; #between different types
my $eps_Spec = 3.5; #max of Specific interactions
my $eps_LAD  = 4.0; #LAD-NE interactions
my $eps_rep  = 3.0; #for pure bead-bead and bead-NE repulsion

# ==========================================================================
# Code hashes for system variables
# --------------------------------
my %radii = ();
my %types = ();
my $counter = 1;
my %masses = ();
my %labels = ();
my %attachments = ();

# ==========================================================================
# Block to load all 6 arms of chromosomes
# ----------------------------------------
# You want to store more than one value for each key in a hash.
# Store an array reference in %hash{$key}, and put the values into that array.
my %lamin = ( '2L'=>\@L2lam,'3L'=>\@L3lam,'2R'=>\@R2lam,'3R'=>\@R3lam,'X'=>\@Xlam,'4'=>\@C4lam );
my %names = ( '2L'=>\@L2,'3L'=>\@L3,'2R'=>\@R2,'3R'=>\@R3,'X'=>\@X, '4'=>\@C4 );
my %order = ( 1 => '2L', 2 => '2R', 3 => '3L', 4 => '3R', 5 => '4',  6 => 'X'); # Order of beads is changed comp to Nick's

my %model = (); # a hash to hold the entire model
#open(OUT,">HET_marks.txt");
my $total_beads = 0;
foreach my $num ( sort { $a <=> $b } keys %order ) {
	my $chromosome = $order{$num};
	my $beadNum = 0; 
	$model{$chromosome} = ();
	foreach my $bead ( @{$names{$chromosome}} ) {
		my @params = split("\t",$bead); 
		chop( @params[5] ) ;  # NEW, to chop off symbol ^M (\r of CR) from the end of last field in @params[5]

		$model{$chromosome}{$beadNum} = ();
		$model{$chromosome}{$beadNum}{'basePairs'} = @params[3]-@params[2]+1;
		$model{$chromosome}{$beadNum}{'from'} = @params[3];
		$model{$chromosome}{$beadNum}{'goto'} = @params[2];
		$model{$chromosome}{$beadNum}{'lamSites'} = count_lamin_binding($lamin{$chromosome},@params[2],@params[3]);
		if ( @params[5] =~ 'HET' ) {$model{$chromosome}{$beadNum}{'lamSites'} = 0;}
		if ( @params[5] =~ 'CEN' ) {$model{$chromosome}{$beadNum}{'lamSites'} = 0;}
#		if ( @params[5] eq 'CEN' ) {   #with     "chop( @params[5] );"  command above

#		$model{$chromosome}{$beadNum}{'radius'} = nearest($roundRadiusDigit,@params[4])/10; # in nm/10
#							TAD radius correction coefficient $corcoeff
		$model{$chromosome}{$beadNum}{'radius'} = nearest($roundRadiusDigit,$corcoeff*@params[4])/10; # NEW, in nm/10
#						$corcoeff  increases the bead radius due to diploid chromosome set
#						Later, the bead radius will be converted to micrometers by additional  /100
#TMP
#print OUT @params[4] .  "\n";
#print OUT $model{$chromosome}{$beadNum}{'radius'} . "\n";
		$model{$chromosome}{$beadNum}{'assignment'} = @params[5];
		$model{$chromosome}{$beadNum}{'nucleosomes'} = $model{$chromosome}{$beadNum}{'basePairs'}/$basePerNucleosome;
# 		Calculation of current bead mass (weight), final weight is in gramm
		$model{$chromosome}{$beadNum}{'weight'} = $model{$chromosome}{$beadNum}{'nucleosomes'} * $NucleosomeDaltons;
		$model{$chromosome}{$beadNum}{'weight'} += $model{$chromosome}{$beadNum}{'basePairs'} * $basePairDaltons;
		$model{$chromosome}{$beadNum}{'weight'} *= $daltonToGram;
		$model{$chromosome}{$beadNum}{'weight'} *= 2;			# NEW diploid chromosomes, mass x2
		$model{$chromosome}{$beadNum}{'weight'} = $model{$chromosome}{$beadNum}{'weight'};
#		if ( @params[5] =~ 'HET' ) {$model{$chromosome}{$beadNum}{'weight'} *= 0.05;}  # NEW, 20x reduction of the HET bead mass
#		The above line (from ONE) was commented out in Making_ALL phase.   Done in 06/21/18 code
		$model{$chromosome}{$beadNum}{'tad'} = @params[0]; 		# NEW TAD ID from the "AlberModelHet.txt" file.
		$model{$chromosome}{$beadNum}{'label'} = tad_label(@params[0]);	# NEW, label for specific TAD-TAD interactions. 
# The value returned by "tad_label()" function above is 0 or > 10 (if 
# Labels from 0 to 5 (TAD type) will be assigned BELOW.
#		print OUT ($model{$chromosome}{$beadNum}{'label'}) . "\n";
# If 0, the 'label' is redefined below 

# ================================================================================================================================
		if ( @params[5] =~ 'Null' && $model{$chromosome}{$beadNum}{'label'} == 0 ) {
			$model{$chromosome}{$beadNum}{'label'} = 0; # NEW, label = 0 to 5 reserved for nonspecific TAD-TAD interactions.
#	 		print OUT ($model{$chromosome}{$beadNum}{'label'}) . "\n";
			}
		if ( @params[5] =~ 'Active' && $model{$chromosome}{$beadNum}{'label'} == 0 ) {
			$model{$chromosome}{$beadNum}{'label'} = 1;            
#			print OUT ($model{$chromosome}{$beadNum}{'label'}) . "\n";
			}	
		if ( @params[5] =~ 'PcG' && $model{$chromosome}{$beadNum}{'label'} == 0 ) {
			$model{$chromosome}{$beadNum}{'label'} = 2;            
#			print OUT ($model{$chromosome}{$beadNum}{'label'}) . "\n";
                        }
		if ( @params[5] =~ 'HP1_centromeric' && $model{$chromosome}{$beadNum}{'label'} == 0 ) {
			$model{$chromosome}{$beadNum}{'label'} = 3;
#			print OUT ($model{$chromosome}{$beadNum}{'label'}) . "\n";
                        }
#		if ( @params[5] =~ 'HET' && $model{$chromosome}{$beadNum}{'label'} == 0 ) {
		if ( @params[5] =~ 'HET') {
			$model{$chromosome}{$beadNum}{'label'} = 4;
#			print OUT ($model{$chromosome}{$beadNum}{'label'}) . "\n";
			}
#print OUT $model{$chromosome}{$beadNum}{'label'} . "\n";
#print OUT "------\n";
# ==================================================================================================================================	
		if ( @params[5] =~ 'CEN' ) {					# NEW
#		if ( @params[5] eq 'CEN' ) {#    with  chop( @params[5] );  command above
			$model{$chromosome}{$beadNum}{'label'} = 5;
			$model{$chromosome}{$beadNum}{'centromere'} = 1;
			} else {
			$model{$chromosome}{$beadNum}{'centromere'} = 0;
			}
#TMP
#print OUT $model{$chromosome}{$beadNum}{'label'} . "\n";
# ==================================================================================================================================

		$model{$chromosome}{$beadNum}{'number'} = $total_beads;		# Numbering ALL beads starting from 0 (0 for VMD?).
		$model{$chromosome}{$beadNum}{'chromosome'} = substr($chromosome,0,1);
#		Extracting the chromosome number, e.g. from 3L - 3, from 2R - 2.

		$model{$chromosome}{$beadNum}{'attachment'} = 0;

# For Pickersgil's scheme the threshhold number of contacts with NE was 7. 
# For now, all HET and CEN beads have {'lamSites'} = 0, thus, they all will have {'attachment'} = 0
		if ( $model{$chromosome}{$beadNum}{'lamSites'} >= 1 ) {
			$model{$chromosome}{$beadNum}{'attachment'} = 1;
			}
#TMP
#print OUT $model{$chromosome}{$beadNum}{'attachment'} . "\n";
#
		if ( $colorByAttach ) { 
			if ( $model{$chromosome}{$beadNum}{'attachment'} ) {
				$model{$chromosome}{$beadNum}{'charge'} = 2; # attachment = 1
				} else {
				$model{$chromosome}{$beadNum}{'charge'} = 1; # attachment = 0
				}
		} else {
			if ($chromosome =~ /2/) { $model{$chromosome}{$beadNum}{'charge'} = 2; }
			if ($chromosome =~ /3/) { $model{$chromosome}{$beadNum}{'charge'} = 3; }
			if ($chromosome =~ /X/) { $model{$chromosome}{$beadNum}{'charge'} = 1; }
			if ($chromosome =~ /4/) { $model{$chromosome}{$beadNum}{'charge'} = 4; }
		}
#TMP
#print OUT "$model{$chromosome}{$beadNum}{'attachment'}  $model{$chromosome}{$beadNum}{'tad'}  \n";
#print OUT $model{$chromosome}{$beadNum}{'charge'} . "\n";
		
		$radii{$model{$chromosome}{$beadNum}{'radius'}}++;		# Hash for radii distribution
		$attachments{$model{$chromosome}{$beadNum}{'attachment'}}++;	# Hash for attachment type distribution
		$labels{$model{$chromosome}{$beadNum}{'label'}}++;		# Hash for label type distribution
		$masses{$total_beads} = $model{$chromosome}{$beadNum}{'weight'};	#Hash for mass disribution, in g here

		
		my $r = $model{$chromosome}{$beadNum}{'radius'}; # Here the bead radius is rounded, in nm/10 units
		my $a = $model{$chromosome}{$beadNum}{'attachment'};
		my $l = $model{$chromosome}{$beadNum}{'label'};

# ================================================================================
# This is a really important step to keep track of the particle type: %types = () 
# ================================================================================	
		$model{$chromosome}{$beadNum}{'type'} = $r . "_" . $a . "_" . $l;
		
		if ( ! exists $types{$r . "_" . $a . "_" . $l} ) {	
			$types{$r . "_" . $a . "_" . $l} = $counter++;
			}
		
		$total_beads++;
		$beadNum++;
	}
}
#close(OUT);

#TMP
#close(OUT);
#print Dumper \%labels;
#print Dumper \%attachments;
#print Dumper \$lamin{'2L'};
#print Dumper \%radii;

	
# ==============================================================
# Not needed with AlberModelHet. All Centromeres (CEN) are now
# explicitly included.
#
# $model{'2R'}{0}{'centromere'} = 1;
# $model{'3R'}{0}{'centromere'} = 1;
# $model{'X'}{max( keys %{$model{'X'}} )}{'centromere'}  = 1;
# ==============================================================

open(OUT,">simulation_params.txt");
print OUT $output_frames . "\n";
print OUT $total_beads . "\n";
#
print OUT ( $model{'X'}{min( keys %{$model{'X'}} )}{'number'}+1 ) . "\n";
print OUT ( $model{'X'}{max( keys %{$model{'X'}} )}{'number'}+1 ) . "\n"; #X centromere (last bead of X)
print OUT ( $model{'X'}{max( keys %{$model{'X'}} )}{'number'}+1 ) . "\n";
#
print OUT ( $model{'2L'}{min( keys %{$model{'2L'}} )}{'number'}+1 ) . "\n";
print OUT ( $model{'2L'}{max( keys %{$model{'2L'}} )}{'number'}+1 ) . "\n"; #2LR centromere (last bead of 2L)
print OUT ( $model{'2R'}{max( keys %{$model{'2R'}} )}{'number'}+1 ) . "\n";
#
print OUT ( $model{'3L'}{min( keys %{$model{'3L'}} )}{'number'}+1 ) . "\n";
print OUT ( $model{'3L'}{max( keys %{$model{'3L'}} )}{'number'}+1 ) . "\n"; #3LR centromere (last bead of 3L)
print OUT ( $model{'3R'}{max( keys %{$model{'3R'}} )}{'number'}+1 ) . "\n";
print OUT $nucleus_radius . "\n";
print OUT int(rand()*1000000) . "\n"; # this is the random seed used to generate the structures
close(OUT);

open(OUT,">simulation_bRadii.txt");
foreach my $num ( sort { $a <=> $b } keys %order ){
	my $chromosome = $order{$num};
	foreach my $bead ( sort { $a <=> $b } keys %{$model{$chromosome}} ) {
		print OUT ($model{$chromosome}{$bead}{'radius'}/100) . "\n";  #Bead radii in micrometers (/10 before and /100 now)
	}
}
print OUT  $nucleolus_radius . "\n";
close(OUT);

# -------------------------------------------------------------------------------------------
# Not needed if to use Alber's model
# ----------------------------------
# if ( $autoGeneratePoints == 0 ) { # this generates a file called strands.txt
#	system("matlab -nodesktop -nosplash < make_structures.m >/dev/null 2> logfile.txt");
#	}
# -------------------------------------------------------------------------------------------


my $minType; # this will hold the time-step limiting bead type
my $c_mass = min(values %masses); # the smallest bead mass in grams
# All masses will be in multiples of this (c_mass) min mass (see "Particle types are redefined" section)


my @radii = sort { $a <=> $b } keys %radii;  # sorting bead radii
my @attachments = sort { $a <=> $b } keys %attachments;
my @labels = sort { $a <=> $b } keys %labels;


# Determination of the type with smallest (radius) bead
my $minRad = @radii[0]/100. ; # This is the radius value (after sorting) for the minType bead
foreach my $type1 ( keys %types ) {
	my @labels1 = split("_",$type1);
	my $r1 = $labels1[0];
	if ( $r1 == @radii[0] ) { $minType = $type1; }  # set the minType variable
	}
	
# ==========================================================================
# PRINTING to tcl file general PARAMETERS and the characteristic units 
# ==========================================================================
print_head();
print_box($box_length);
print_characteristic_units($c_mass); # NEW, depends on $c_mass value, not a constant 77MDa bead mass as before
# $c_mass is converted from g to kg inside this subroutine for PRINTOUT ONLY.
# Inside this perl code it is in grams. Later, when particles will be defined, the all masses will be converted
# to relative units using this #c_mass - the smallest bead mass (see $c_mass = min() function abobe).
print_espresso_units();
#print_interac_params($eps_Null,$eps_Act,$eps_PcG,$eps_HP1,$eps_gen,$eps_Spec,$eps_LAD,$eps_rep);
#Function above is replaced with a direct inserting of the above parameters.
# ==========================================================================================
	

# Give the nucleolus and nuclear envelope (NE) type and ID
my $nucleolus_type = max(values %types)+1;
my $nucleolus_numb = $total_beads++;
my $nEnvelope_type = max(values %types)+2;
# This type is used for the COM center. Actual NE type is 0. ??? Check !!!
my $nEnvelope_numb = $total_beads++;
my $extra_type = max(values %types)+3;
my $extra_numb = $total_beads++;

	
# ==========================================================================================
# PRINTING all INTERACTION and BOND PARAMETERS (LJ, harmonic springs) needed for simulation
# ==========================================================================================
	print "# interaction parameters for TAD-TAD interactions \n";
# Below, the lines are commented out as well as   "print_interac_params()" function above.
#	my $epsN = "[expr {\$epsN*\$kb*\$temp}]";
#	my $epsA = "[expr {\$epsA*\$kb*\$temp}]";
#	my $epsP = "[expr {\$epsP*\$kb*\$temp}]";
#	my $epsH = "[expr {\$epsH*\$kb*\$temp}]";
#	my $epsG = "[expr {\$epsG*\$kb*\$temp}]";
#	my $epsS = "[expr {\$epsS*\$kb*\$temp}]";
#	my $epsL = "[expr {\$epsL*\$kb*\$temp}]";
	my $epsN = "[expr {" . $eps_Null . "*\$kb*\$temp}]";
	my $epsA = "[expr {" . $eps_Act . "*\$kb*\$temp}]";
	my $epsP = "[expr {" . $eps_PcG . "*\$kb*\$temp}]";
	my $epsH = "[expr {" . $eps_HP1 . "*\$kb*\$temp}]";
	my $epsG = "[expr {" . $eps_gen . "*\$kb*\$temp}]";
	my $epsS = "[expr {" . $eps_Spec . "*\$kb*\$temp}]";
	my $epsL = "[expr {" . $eps_LAD . "*\$kb*\$temp}]";
	my $epsR = "[expr {" . $eps_rep . "*\$kb*\$temp}]";
	my $epsilon = "[expr {3.0*\$kb*\$temp}]";
# -----------------------------------------------------------
# printing LJ epsilon values for different interactions
	print "set lj_epsN\t" . $epsN . "\n";
	print "set lj_epsA\t" . $epsA . "\n";
	print "set lj_epsP\t" . $epsP . "\n";
	print "set lj_epsH\t" . $epsH . "\n";
	print "set lj_epsG\t" . $epsG . "\n";
	print "set lj_epsS\t" . $epsS . "\n";
	print "set lj_epsL\t" . $epsL . "\n";
	print "set lj_epsR\t" . $epsR . "\n\n";
# -----------------------------------------------------------

foreach my $type1 ( keys %types ) {
	my @labels1 = split("_",$type1);
	my $r1 = $labels1[0];
	foreach my $type2 ( keys %types ) {
		my @labels2 = split("_",$type2);
		my $r2 = $labels2[0];
		my $bondLength = ($r1+$r2)/100.; # Nick's had factor 2 - "((2*($r1+$r2))" -presumably for the warm-up stage
#	Finally, above the radii are devided by 100 (in addition to previous /10) to convert from nm to micrometer units.
		my $sigma = $bondLength/1.12246204831;
		my $cut = "\$b_length_" . $type1 . "_" . $type2;
# ------------- printing other LJ and bond spring constants parameters ------------------------------------------------
		print "# interactions for radii $type1 and $type2\n";
		print "set b_length_" . $type1 . "_" . $type2 . "\t" . $bondLength . "\n";
		print "set lj_" . $type1 . "_" . $type2 . "_sig\t" . $sigma . "\n";
		print "set lj_" . $type1 . "_" . $type2 . "_eps\t" . $epsilon . "\n"; #Used in Stage II. Not anymore, with the introduction of $epsG and $epsR, see Comm. l.715
		print "set lj_" . $type1 . "_" . $type2 . "_cut\t" . $cut . "\n";
		print "set lj_" . $type1 . "_" . $type2 . "_shift\tauto\n";
		print "set k_spring_" . $type1 . "_" . $type2 . "\t\t[expr {(5.0*\$kb*\$temp)/pow(\$b_length_" . $type1 . "_" . $type2 . ",2)}]\n";
		print "set k_spring_" . $type1 . "_" . $type2 . "_strong" . "\t[expr {(10.0*\$kb*\$temp)/pow(\$b_length_" . $type1 . "_" . $type2 . ",2)}]\n\n";
#		Note:   5kT/L**2    and    10kT/L**2 (_strong)   are set here for the bond constants 
	}
}
	
# =========================================================================================
# PRINTING to tcl file model PARAMETERS
# =========================================================================================
print " \n";
print_time_steps($minType); # smallest bead Type determines time step
print " \n";
print "set nucleus_radius " . $nucleus_radius . "\n";
print "set minRad " . $minRad . "\n";
print "constraint sphere center [expr \$box_l/2.0] [expr \$box_l/2.0] [expr \$box_l/2.0] radius [expr \$nucleus_radius + \$minRad] direction -1 type 0 penetrable 0\n";
# Constraint_sphere_radius is [$nucleus_radius + $minRAD]. The addition of $minRad is used.
print " \n";
# ==========================================================================================


# =========================================================================================
# BEAD-BEAD harmonic bonds.
# =========================================================================================
# NEW: Single loop over the beads in each chromosome, not the two loops over the bead types. 
# 4 times smaller output tcl code. Only needed interactions are included
# -----------------------------------------------------------------------------------------
print " \n";
print "\# Nearest bead-bead harmonic bond interaction types\n";
my %harmonic_hash = ();
my $harmonic_type = 1;
foreach my $num ( sort { $a <=> $b } keys %order ) {
	my $chromosome = $order{$num};
	my $maxIndex = max keys %{$model{$chromosome}};
#       my $maxIndex = max(keys %{@model{$chromosome}});
# -----------------------------------------------------------------------------------------
# NEW in making_ALL script. 
# To describe the "bond" between chromosome ARMS, and the corresponding $harmonic_hash
	my $numPlus = $num + 1;
	my $chromPlus = $chromosome;
	if ($numPlus <= 6 ) {
		$chromPlus = $order{$numPlus}; # Next chromosome from the hash %order{}
	}
# ------ loop through the beads in a given chromosome -------------------------------------
	foreach my $bead ( sort { $a <=> $b } keys %{$model{$chromosome}} ) {	
		my $next = $bead + 1;
# Types counting:           $types{$r . "_" . $a . "_" . $l} = $counter++;
# Fore reference only:      $model{$chromosome}{$beadNum}{'type'} = $r . "_" . $a . "_" . $l;
                my $type1 = $model{$chromosome}{$bead}{'type'};

		if ( $next <= $maxIndex ) {
#		if ( exists $model{$chromosome}{$next}{'type'} ) { # THIS line produce extra empty elelment of the hash !!!!!
			my $type2 = $model{$chromosome}{$next}{'type'};
		print "inter " . $harmonic_type . " harmonic \$k_spring_" . $type1 . "_" . $type2 . " \$b_length_" . $type1 . "_" . $type2 . "\n";
			$harmonic_hash{$type1 . "_" . $type2} = $harmonic_type;
			$harmonic_hash{$type2 . "_" . $type1} = $harmonic_type;
			$harmonic_type++;
		print "inter " . $harmonic_type . " harmonic \$k_spring_" . $type1 . "_" . $type2 . "_strong" . " \$b_length_" . $type1 . "_" . $type2 . "\n";
			$harmonic_hash{$type1 . "_" . $type2 . "_strong"} = $harmonic_type;
			$harmonic_hash{$type2 . "_" . $type1 . "_strong"} = $harmonic_type;
			$harmonic_type++;
		} else {
# $bead counter exceeds the number of beads in the current chromosome. Switching to $chromPlus first bead - "0"
			my $type2 = $model{$chromPlus}{'0'}{'type'};
		print "inter " . $harmonic_type . " harmonic \$k_spring_" . $type1 . "_" . $type2 . " \$b_length_" . $type1 . "_" . $type2 . "\n";
			$harmonic_hash{$type1 . "_" . $type2} = $harmonic_type;
			$harmonic_hash{$type2 . "_" . $type1} = $harmonic_type;
			$harmonic_type++;
		print "inter " . $harmonic_type . " harmonic \$k_spring_" . $type1 . "_" . $type2 . "_strong" . " \$b_length_" . $type1 . "_" . $type2 . "\n";
			$harmonic_hash{$type1 . "_" . $type2 . "_strong"} = $harmonic_type;
			$harmonic_hash{$type2 . "_" . $type1 . "_strong"} = $harmonic_type;
			$harmonic_type++;
		}
	}
}
# ------------------------------------------------------------------------------------------


# ==========================================================================================
# PARAMETERS of BEADS interactions with the COMmon center to COMPRESS chromosomes (stage I).
# These are used in the system PREPARATION stage ONLY (stage I).
# ==========================================================================================
# NOT NEEDED HERE, in the pure simulation stage III,  "sim_ALL..." script
# ------------------------------------------------------------------------------------------
# NK: $bondLength has to be set to 1.00  here. 
# $bondLength=1.125 ($nucleus_radius/2) leads to instability at warm-up
#
# See also the set of COM-related interactions two block down 
# where the INTERACTIONS using these parameters are set.
=pod
	print "# PARAMETERS for interactions of beads with the COM center for chromosome compression\n";
	foreach my $type1 ( keys %types ) {
	my @labels1 = split("_",$type1);
	my $r1 = $labels1[0];
	my $bondLength = 0.99;
	my $weight = .5; # AMPLITUDE of the spring constant, bead-COM harmonic bond. 0.5 was used in test runs
	print "set b_length_" . $type1 . "_COM\t" . $bondLength . "\n";
	print "set k_spring_" . $type1 . "_COM\t[expr {($weight*\$kb*\$temp)/pow(\$b_length_" . $type1 . "_COM,2)}]\n\n";
}

# ==========================================================================================
# Harmonic INTERACTIONS with COM center for condesing each chromosome.
# ==========================================================================================
	print " \n";
	print "\# Bead-COM center harmonic interactions\n";
# Harmonic_type counting continues
	foreach my $type1 ( keys %types ) {
		my @labels1 = split("_",$type1);
		my $r1 = $labels1[0];
		print "inter " . $harmonic_type . " harmonic \$k_spring_" . $type1 . "_COM \$b_length_" . $type1 . "_COM\n";
		$harmonic_hash{$type1 . "_COM"} = $harmonic_type;
		$harmonic_type++;
	}
=cut
# ------------------------------------------------------------------------------------------


# ==========================================================================================
# Bead-Nucleolus interaction PARAMETERS. 
# ==========================================================================================
print "\nset nucleolus_radius $nucleolus_radius\n";
my $epslus = "[expr {3.0*\$kb*\$temp}]"; # Setting the "epsilon" for bead-nucleolus interactions
my $epslus_strong = "[expr {6.0*\$kb*\$temp}]"; # Strong version for HET-nucleolus interactions
print "set epslus $epslus\n";
print "set epslus_strong $epslus_strong\n\n";

# ==========================================================================================
# REPULSIVE Bead-Nucleolus interactions. ATTRACTIVE for HET beads (label = 4)
# ==========================================================================================
print "\# Repulsive (TADs) and Attracive (HET beads) LJ interactions with Nucleolus\n";
foreach my $type1 ( keys %types ) {
	my @labels1 = split("_",$type1);
	my $r1 = $labels1[0];
	my $tlb1 = $labels1[2];
	if ( $tlb1 == '4' ) { # field #3 ([2],label) = 4  - HET beads; ATTRACTIVE LJ interaction with $epslus_strong = 6 kT.
	printf("inter %d %d lennard-jones \$epslus_strong [expr {(\$nucleolus_radius+%0.2f)/1.12246204831}] 4.0 0.0\n",$types{$type1},$nucleolus_type,$r1/100);
	}  else { # all other beads - REPULSIVE LJ interaction.                       r_cut = ($nucleolus_radius + $r1/100).   c_shift = auto.
	printf("inter %d %d lennard-jones \$epslus [expr {(\$nucleolus_radius+%0.2f)/1.12246204831}] [expr {\$nucleolus_radius+%0.2f}] auto\n",$types{$type1},$nucleolus_type,$r1/100,$r1/100);
# bead radii are finaly converted to micrometers here using factors 1/100.
	}
}
# ------------------------------------------------------------------------------------------


###########################################################################################################################
# Generate the positions of the particles
my @points = ();
# The order of coordinates below should correspond to the revers order in hash %order. pop() is used to exract last element.
my @positions = ("51 51 50","50 50 54","51 49.5 50","50 49 50","49 51 50","49 49.5 50");
#                    X          4          3R           3L         2R          2L

# The above coordinates define CIS configuration of chromosomes [Fig.2A in Hochstrasser 1986, or Fig.3 (7) in Kinney 2014] 
###########################################################################################################################
if ( $autoGeneratePoints == 0 ) { # MATLAB point generation is used if == 0.  strands.txt file is needed here.
	chomp( @points = `cat strands.txt`);
	for my $i ( 0..$#points ) { $points[$i] =~ s/ /\t/g; }
	} else {
		foreach my $chromosome ( sort { $a <=> $b } keys %order ) {
# Here, $chromosome are 1,2,3,4,5,6
			my @a = generate_chain($model{$order{$chromosome}});
			shift_chain(\@a,subtract(pop(@positions),center_of_mass(\@a,$model{$order{$chromosome}})));
			if ( $order{$chromosome} =~ /R/ ) { flip(\@a); }
			foreach ( @a ) { push @points,$_; }
			}
	for my $i ( 0..$#points ) { $points[$i] =~ s/ /\t/g; }
	}
# -------------------------------------------------------------------------------------------------------------------------

# ================================================================================================================================
# Reading ALL particles coordinates from the file "config_ALL" file generated by the Making_ALL script
# ================================================================================================================================
print " \n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "\# Define the points (beads). Reading from the \"config_ALL\" file.\n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";

# ------------------------------------------------------------
print "set f [open \"config_ALL\" \"r\"]\n";
print "while { [blockfile \$f read auto] != \"eof\" } {}\n";
print "close \$f \n";
# ------------------------------------------------------------
print " \n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "\# Particle types are redefined here, can be different from what\n";
print "\# are in input \"config_ALL\" file. \n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
# Block to specify beads (including HET and CEN) from ALL 6 chromosome arms and two extra beads (Nucleolus and NE center).
# Modified !!!  Coordinates are excluded (see commented off line below replaced with NO pos parameter.
foreach my $num ( sort { $a <=> $b } keys %order ) {
	my $chromosome = $order{$num};
	foreach my $bead ( sort { $a <=> $b } keys %{$model{$chromosome}} ) {
		my @point = split("\t",@points[$model{$chromosome}{$bead}{'number'}]);
		my $m = $model{$chromosome}{$bead}{'weight'}/$c_mass; # Normalization of bead mass onto the smallest bead mass (in grams here).
		printf("part %s type %d mass %5.3f q %d\n",$model{$chromosome}{$bead}{'number'},$types{$model{$chromosome}{$bead}{'type'}},$m,$model{$chromosome}{$bead}{'charge'});
#		Typing positions (pos) is excluded above. They remain the same as in config_ALL file.
#		printf("part %s pos %5.3f %5.3f %5.3f type %d mass %5.3f q %d\n",$model{$chromosome}{$bead}{'number'},@point[0],@point[1],@point[2],$types{$model{$chromosome}{$bead}{'type'}},$m,$model{$chromosome}{$bead}{'charge'}); # This one is with printing [pos] xyz coordinates.
		$model{$chromosome}{$bead}{'point'} = @point[0] . " " . @point[1] . " " . @point[2];
		if ( $bead > 0 ) { # this loop anticipates the bonds which are set next
			if ( $model{$chromosome}{$bead-1}{'chromosome'} == $model{$chromosome}{$bead}{'chromosome'} ) {
				$model{$chromosome}{$bead-1}{'bond'} = $model{$chromosome}{$bead}{'number'};
			}
		}
	}
}
# ----------------------------------------------------------------------------------------------------------------------------------
# Two extra particles: Nucleolus and Nuclear Envelope (NE) center. 
# (x,y,z) coordinates are not actually need here. They are read from "config_ALL" file as well as all other beads coordinates.
print "\n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "\# Extra particles: Nucleolus and nuclear Envelope (NE) center\n";
# Redefined here
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "part " . $nucleolus_numb . " pos [expr \$box_l/2.0] [expr \$box_l/2.0] [expr (\$box_l/2.0) + " . ($nucleus_radius/2) . "] type " . $nucleolus_type . " mass 1 q 0 fix 1 1 1\n";
print "part " . $nEnvelope_numb . " pos [expr \$box_l/2.0] [expr \$box_l/2.0] [expr \$box_l/2.0] type " . $nEnvelope_type . " mass 1 q 0 fix 1 1 1\n\n";
#print "part " . $extra_numb . " pos [expr \$box_l/2.0] [expr \$box_l/2.0] [expr \$box_l/2.0] type " . $extra_type . " mass 1 q 0 fix 1 1 1\n\n";
# Commented out (temporaly).
# =================================================================================================================================


# =====================================================================
# At this point ALL the particles and their coordinates are defined !!!

# Particle ID "$nEnvelope_numb" is a dummy center NOW - no interactions.
# Interactions with NE are defined as a CONSTRAINT which has type 0 and 
# is defined as "constraint sphere". 
# Can be replaced by uncommenting  $extra_numb  bead.
# =====================================================================
# print "part " . $nEnvelope_numb . " delete\n";
# print "kill_particle_motion\n\n";



# =================================================================================================================================
# And finally, creating NEW bead-bead BONDS with about 2 times stronger bond constant (index "_strong")
# =================================================================================================================================
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "\# Place the bead-bead bonds. Connect the arms\n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "\# Creating strong (10 kT vs 5kT) bonds \n";
# ===========================================================================
# NEW: centromere (CEN) beads are now LAST beads of 2L, 3L and X chromosomes.
# ===========================================================================
# Connection scheme is redefined
# Bead-bead and Arms connecting bonds can be different

foreach my $num ( sort { $a <=> $b } keys %order ) {
	my $chromosome = $order{$num};
	foreach my $bead ( sort { $a <=> $b } keys %{$model{$chromosome}} ) {
		my $next = $bead + 1;
		if ( ( exists $model{$chromosome}{$bead}{'bond'} ) & ( exists $model{$chromosome}{$next} ) ) {
			my $harmonic_key = $model{$chromosome}{$bead}{'type'} . "_" . $model{$chromosome}{$next}{'type'} . "_strong";
			printf("part %d bond %d %d\n",$model{$chromosome}{$bead}{'number'},$harmonic_hash{$harmonic_key},$model{$chromosome}{$bead}{'bond'});
		}
		if ( $model{$chromosome}{$bead}{'centromere'} == 1 ) {
			if ( $chromosome =~ '2L' ) {
			my $harmonic_key = $model{$chromosome}{$bead}{'type'} . "_" . $model{'2R'}{'0'}{'type'} . "_strong";
			printf("part %d bond %d %d\n",$model{$chromosome}{$bead}{'number'},$harmonic_hash{$harmonic_key},$model{'2R'}{'0'}{'number'});
			}
			if ( $chromosome =~ '3L' ) {
			my $harmonic_key = $model{$chromosome}{$bead}{'type'} . "_" . $model{'3R'}{'0'}{'type'} . "_strong";
			printf("part %d bond %d %d\n",$model{$chromosome}{$bead}{'number'},$harmonic_hash{$harmonic_key},$model{'3R'}{'0'}{'number'});
			}
		}
	}
}
# --------------------------------------------------------------------------------------------------------------------------------



# ====================================================================================
# INTERACTIONS with with the NE SPHERE, using spherical constraint defined in
# print_constraint_sphere() function which produces "constraint sphere center" command
# [Modified now to just a command with a changed radius].
# ====================================================================================
# REPULSIVE only HERE.
print " \n";
print "\# Bead-NE (constraint sphere, type 0) REPULSIVE LJ interactions\n";
foreach my $type1 ( keys %types ) {
	my @labels1 = split("_",$type1);
	my $r1 = $labels1[0];
        if ( $envelope_attachments_on ) {
		my $a1 = $labels1[1];
		if ( $a1 == 1 ) { next; } # Beads with $a1 == 1 (LADs) are not included in the REPULSIVE interactions.
		}
# print "inter 0 " . $types{$type1} . " lennard-jones \$lj_epsL  [expr \$lj_" . $type1 . "_" . $minType . "_sig] [expr \$lj_" . $type1 . "_" . $minType . "_cut] auto \n";
# 1) Bad when $lj_epsL ~ 0kT; 2) Should not depend when $lj_epsL is changed.
print "inter 0 " . $types{$type1} . " lennard-jones \$lj_epsR  [expr \$lj_" . $type1 . "_" . $minType . "_sig] [expr \$lj_" . $type1 . "_" . $minType . "_cut] auto \n";
	}	

# ATTRACTIVE (LAD) interactions with the NE
print " \n";
print "\# Bead-NE (LAD-NE, constraint sphere, type 0) ATTRACTIVE LJ-COS interactions\n";
if ( $envelope_attachments_on ) {
foreach my $type1 ( keys %types ) {
	my @labels1 = split("_",$type1);
	my $a1 = $labels1[1];
	if ( $a1 == 1 ) {
	print "inter 0 " . $types{$type1} . " lj-cos \$lj_epsL [expr \$lj_" . $type1 . "_" . $minType . "_sig] [expr (1.2*\$b_length_" . $type1 . "_" . $minType . ")] 0\n";
		}
	}
}
# =============================================================================

=pod
# ========================================================================================
# Output the REPULSIVE bead-bead (TAD-TAD) interactions: nonspecific TAD-TAD interactions
# ==+=====================================================================================
# The below  REPULSIVE bead-bead interactions are commented OUT because the introduced
# below NONspecific small (0.1 to 1kT) attraction between ALL epigenetic types.
# If some types are excluded, the above repulsion for that types has to be REINTRODUCED!!!
# ========================================================================================
print "\# Bead-bead (TAD-TAD) repulsive LJ interactions\n";
foreach my $type1 ( keys %types ) {
	my @lab1 = split("_",$type1);
	my $tlb1 = $lab1[2];
	my $r1 = $lab1[0];
	foreach my $type2 ( keys %types ) {
		my @lab2 = split("_",$type2);
		my $tlb2 = $lab2[2];
		my $r2 = $lab2[0];
		if ( $types{$type1} > $types{$type2} ) { next; } # New in the latest.
#		if ( $r2 > $r1 ) { next; }   # Alternative to type1, type2 order; 251 lines more
		if ( $tads_on ) {
			if ( $tlb1 < 6 && $tlb2 < 6) {
print "inter $types{$type1} $types{$type2} lennard-jones \$lj_" . $type1 . "_" . $type2 . "_eps \$lj_" . $type1 . "_" . $type2 . "_sig \$lj_" . $type1 . "_" . $type2 . "_cut \$lj_" . $type1 . "_" . $type2 . "_shift\n";
			}
		} else { # The same as above but without restriction by the TAD label ($tlb).
print "inter $types{$type1} $types{$type2} lennard-jones \$lj_" . $type1 . "_" . $type2 . "_eps \$lj_" . $type1 . "_" . $type2 . "_sig \$lj_" . $type1 . "_" . $type2 . "_cut \$lj_" . $type1 . "_" . $type2 . "_shift\n";
		}
	}
}

# The above REPULSIVE bead-bead interactions are commented out because the introduced below
# NONspecific small (0.1 to 1kT) attraction ($epsG) between ALL epigenetic types (see l.744).
# If some types are excluded, the above repulsion for that types has to be REINTRODUCED!!!
# ========================================================================================
=cut

# ==============================================================================
# Nonspecific TAD-TAD interactions (now between ALL epigenetic types (0 to 3) ).
# ==============================================================================
print " \n";
print "\# Non-specific bead-bead (TAD-TAD) attractive LJ interactions\n";
print "\# All cross-types are included here\n";
foreach my $type1 ( keys %types ) {
	my @lab1 = split("_",$type1);
	my $tlb1 = $lab1[2];
	my $r1 = $lab1[0];
	foreach my $type2 ( keys %types ) {
		my @lab2 = split("_",$type2);
		my $tlb2 = $lab2[2];
		my $r2 = $lab2[0];
# --------------------------------------------------------------
		if ( $types{$type2} > $types{$type1} ) { next; }  #To skip half of the interaction matrix
		if ( $tads_on ) {
			if ( $tlb1 == 0 && $tlb1 == $tlb2 ) { next; } # New of June 20, 2019
			if ( $tlb1 == 1 && $tlb1 == $tlb2 ) { next; } # New of June 20, 2019
			if ( $tlb1 == 2 && $tlb1 == $tlb2 ) { next; } # New of June 20, 2019
			if ( $tlb1 == 3 && $tlb1 == $tlb2 ) { next; } # New of Apr 24, 2019: to exclude HP1-HP1 interactions, 
			if ( $tlb1  > 9 && $tlb1 == $tlb2 ) { next; } # Specific pairs
# (0) Small attraction ($lj_epsG ~0.1 kT) between all beads not specified above
  print "inter $types{$type1} $types{$type2} lj-cos \$lj_epsG [expr \$lj_" . $type1 . "_" . $type2 . "_sig] [expr (1.2*\$b_length_" . $type1 . "_" . $type2 . ")] 0\n";
# print "inter $types{$type1} $types{$type2} lennard-jones \$lj_espG \$lj_" . $type1 . "_" . $type2 . "_sig 0.6 0.0\n";
			} else {  # for $tads_on = 0  (no TAD-TAD attraction, REPULSION only)
# (1) Pure repulsion for $tads_on = 0 - shifted lennard-jones
  print "inter $types{$type1} $types{$type2} lennard-jones \$lj_epsR \$lj_" . $type1 . "_" . $type2 . "_sig \$lj_" . $type1 . "_" . $type2 . "_cut \$lj_" . $type1 . "_" . $type2 . "_shift\n"
# --------------------------------------------------------------
# (0) Small well attracton (~0.1 kT) lj-cos (not lennard-jones)
# print "inter $types{$type1} $types{$type2} lj-cos \$lj_epsG [expr \$lj_" . $type1 . "_" . $type2 . "_sig] [expr (1.2*\$b_length_" . $type1 . "_" . $type2 . ")] 0\n";
# -------------------------------------------------------------------------
# (0-alternative) Small well attracton (~0.1 kT) lennard-jones (not lj-cos)
# print "inter $types{$type1} $types{$type2} lennard-jones \$lj_epsG \$lj_" . $type1 . "_" . $type2 . "_sig 0.6 0.0\n";
# -------------------------------------------------------------------------
		}
	}
}
# ================================================================================

# ===============================================================================
# Additional TAD-TAD, specific and HP1-HP1 interactions here
# ===============================================================================
# NEW
print " \n";
print "\# The same type and Specific (selective Sexton based) TAD-TAD variable attractive LJ interactions\n";
if ( $tads_on ) {
	foreach my $type1 ( keys %types ) {
		my @lab1 = split("_",$type1);
		my $tlb1 = $lab1[2];
		my $r1 = $lab1[0];
		foreach my $type2 ( keys %types ) {
			my @lab2 = split("_",$type2);
			my $tlb2 = $lab2[2];
			my $r2 = $lab2[0];
# -----------------------------------------------------------------------
			if ( $types{$type2} > $types{$type1} ) { next; }
			if ( $tlb1 == 0 && $tlb1 == $tlb2 ) { #Null-Null 
	print "inter $types{$type1} $types{$type2} lj-cos \$lj_epsN  [expr \$lj_" . $type1 . "_" . $type2 . "_sig] [expr (1.2*\$b_length_" . $type1 . "_" . $type2 . ")] 0\n";
#	print "inter $types{$type1} $types{$type2} lennard-jones \$lj_epsN  \$lj_" . $type1 . "_" . $type2 . "_sig  0.6  0.0\n"
			}
                        if ( $tlb1 == 1 && $tlb1 == $tlb2 ) { #Active-Active
	print "inter $types{$type1} $types{$type2} lj-cos \$lj_epsA  [expr \$lj_" . $type1 . "_" . $type2 . "_sig] [expr (1.2*\$b_length_" . $type1 . "_" . $type2 . ")] 0\n";
#	print "inter $types{$type1} $types{$type2} lennard-jones \$lj_epsA  \$lj_" . $type1 . "_" . $type2 . "_sig  0.6  0.0\n"
			}
                        if ( $tlb1 == 2 && $tlb1 == $tlb2 ) { #PcG-PcG
	print "inter $types{$type1} $types{$type2} lj-cos \$lj_epsP  [expr \$lj_" . $type1 . "_" . $type2 . "_sig] [expr (1.2*\$b_length_" . $type1 . "_" . $type2 . ")] 0\n";
#	print "inter $types{$type1} $types{$type2} lennard-jones \$lj_epsP \$lj_" . $type1 . "_" . $type2 . "_sig  0.6  0.0\n"
			}
			if ( $tlb1 == 3 && $tlb1 == $tlb2 ) {
	print "inter $types{$type1} $types{$type2} lj-cos \$lj_epsH  [expr \$lj_" . $type1 . "_" . $type2 . "_sig] [expr (1.2*\$b_length_" . $type1 . "_" . $type2 . ")] 0\n";
#       print "inter $types{$type1} $types{$type2} lennard-jones \$lj_epsH  \$lj_" . $type1 . "_" . $type2 . "_sig  0.6  0.0\n";
			}
			if ( $tlb1 > 9 && $tlb1 == $tlb2 ) { #Labels for selective TAD-TAD start from 11. See tad_label() !!!
	print "inter $types{$type1} $types{$type2} lennard-jones [expr {-1.18*((($tlb1 - 11.)/267.)**0.2) + \$lj_epsS}] \$lj_" . $type1 . "_" . $type2 . "_sig 0.6 0.0\n"; 
# The strongest interaction is $lj_epsS.  With increasing $tbl, the interactions are reduced by up to 1.18kT.
			}
		}
	}
  }
# ===============================================================================


# ===========================================
# Writing trajectory file header
# ===========================================
my $ptcount = 1;

print " \n";
print "\# Starting trajectory file\n";

print "writevsf \$traj_file radius {\n";
# =========================================================
foreach my $type1 ( keys %types ) {
        my @lab1 = split("_",$type1);
        my $r1 = $lab1[0]/100;
        print " " . $types{$type1} . " " . $r1 . "\n";
        }
# =========================================================
=pod
# Adding  "type auto" lines to specify beads' radii
foreach my $type1 ( keys %types ) {
	if ( $ptcount < $nucleolus_type ) {
		print " " . $ptcount . " auto\n";
		$ptcount++;
		}
	}
=cut
# =========================================================
print " " . $nucleolus_type . " " . $nucleolus_radius . "\n";
print " " . $nEnvelope_type . " 0.1\n";
print "}\n";
if ( $output_warmup == 1 ) { print "writevcf \$traj_file folded\n"; }


# ===============================================	
# Final integration. Trajectory generation.
# ===============================================
print "\n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "\# Simulation integration\n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";

print " \n";
print "set gamma\t[expr {0.01/\$sim_time}]\n";
print "thermostat langevin \$temp \$gamma\n";
# Check $output_frame at the begining of this script

print "set int_steps $steps_per_frame\n"; # 1K now
print "set int_n_times  $output_frames\n"; # 40K now
print "puts \"Starting simulation integration: outputting [expr \$int_n_times*\$int_steps] steps\\n\"\n";
print "puts \"gamma set to \$gamma\\n\"\n";
print "puts \"temp set to \$temp\\n\"\n";
print "puts \"Defined Interactions:\\n[inter]\\n\"\n";
print "for {set i 0} { \$i < \$int_n_times } { incr i} {\n";
print "    integrate \$int_steps\n";
print "    writevcf \$traj_file folded\n";
print "    set act_min_dist [analyze mindist]\n";
print "    puts -nonewline \"integration step [expr \$i+1] out of \$int_n_times |  min dist = \$act_min_dist\\r\"\n";
print "}\n";
print "puts \"\\n\\nFinished simulation integration\\n\"\n";


print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";
print "\# End of simulation\n";
print "\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\#\n";

print "close \$traj_file\n";
print "# terminate program\n";

# Writing final system configuration
print " \n";
print "set f [open \"config_ALL_end\" \"w\"]\n";
print "blockfile \$f write tclvariable {box_l}\n";
print "blockfile \$f write variable box_l\n";
print "blockfile \$f write particles {id pos type}\n";
print "close \$f \n";

print "puts \"\\n\\n===================================Finished====================================\\n\\n\"\n";
print "exit\n";

# ==================================================================================================================
# Functions and Subs
# ==================================================================================================================

sub calc_point {
# ---------------
	my @from = split("\t",@_[1]);
	my @goto = split("\t",@_[0]);
	my @star = split(" ",@_[4]);
	my $bond = (@_[2] + @_[3])/100;
	my $x = @goto[0]-@from[0];
	my $y = @goto[1]-@from[1];
	my $z = @goto[2]-@from[2];
	if (abs($x) > 0) { $x = $x / abs($x); }
	if (abs($y) > 0) { $y = $y / abs($y); }
	if (abs($z) > 0) { $z = $z / abs($z); }
	my @point = ($x*$bond+@star[0],$y*$bond+@star[1],$z*$bond+@star[2]);
	return @point;
	}

sub print_constraint_sphere {
# ----------------------------
	print "\n\nconstraint sphere center [expr \$box_l/2.0] [expr \$box_l/2.0] [expr \$box_l/2.0] radius [expr \$box_l/2.0] direction -1 type 0 penetrable 0\n\n";
	}

sub print_characteristic_units {
# ------------------------------
	print "# characteristic units: c_mass - smallest bead mass in kg, c_length - in meters, c_time - in s, c_kb in SI units\n";
#       print "set c_mass\t[expr 127861496.840000*pow(10,-27)]\n"; # Old Nick's 77MDa bead mass in kg
	my $c_mass = @_[0] * 0.001; # Smallest bead mass, converted from g to kg
	print "set c_mass\t" . $c_mass . "\n"; 
	print "set c_length\t[expr 1.000000*pow(10,-6)]\n";
        print "set c_kb\t[expr 1.3806488*pow(10,-23)]\n";
        print "set c_energy\t[expr 298.0*\$c_kb]\n";
	my $c_energy =  298.0* 1.3806488e-23;
	print "#   c_energy = " . $c_energy . " J  [kg*m^2/s^2]  = 1kT\n";
        print "set c_time\t[expr {\$c_length*sqrt(\$c_mass/\$c_energy)}]\n";
	my $c_time = 1.00000e-6*sqrt($c_mass/$c_energy);
	print "#   c_time = " . $c_time . " s \n\n";
	}

sub print_espresso_units {
# ------------------------------
	print "# set the Espresso unitless parameters\n";
	print "set kb\t\t1.0\n";
	print "set mass\t1.0 ; # in c_mass units\n";
#	print "set energy_at_one_kelvin\t[expr (\$c_kb*\$c_time*\$c_time)/(\$c_mass*\$c_length*\$c_length)]\n";
#	print "set temp\t[expr 298*\$energy_at_one_kelvin]\n\n";
        print "set temp\t1.0 ; # in c_energy = 1kT units\n\n";
	}

sub print_interac_params{
# ------------------------
	my $epsN = @_[0];
	my $epsA = @_[1];
	my $epsP = @_[2];
	my $epsH = @_[3];
	my $epsG = @_[4];
	my $epsS = @_[5];
	my $epsL = @_[6];
	my $epsR = @_[7];
	print "# set bead-bead interaction parameters\n";
	print "set epsN\t" . $epsN . "\n";
	print "set epsA\t" . $epsA . "\n";
	print "set epsP\t" . $epsP . "\n";
	print "set epsH\t" . $epsH . "\n";
	print "set epsG\t" . $epsG . "\n";
	print "set epsS\t" . $epsS . "\n";
	print "set epsL\t" . $epsL . "\n";
	print "set epsR\t" . $epsR . "\n\n";
	}


sub print_time_steps {
# ---------------------
	my $type = @_[0];
	my @labels1 = split("_",$type);
	my $radius = $labels1[0];
	print "# set time steps\n";
	print "set time_step\t[expr {0.01*\$lj_" . $type . "_" . $type . "_sig*sqrt(\$mass/([expr {3.0*\$kb*\$temp}]))}]\n";
	print "set sim_time\t[expr {\$lj_" . $type . "_" . $type . "_sig*sqrt(\$mass/([expr {3.0*\$kb*\$temp}]))}]\n";
	print "set gamma\t[expr {0.01/\$sim_time}]\n";
	print "puts \"\\n\\nparameters set with gamma = \$gamma\\n\\n\"\n";
	print "puts \"\\n\\ntime_step value = \$time_step\\n\\n\"\n";
	print "setmd time_step\t\$time_step\n";
	print "setmd skin\t.1\n";
	print "thermostat langevin \$temp \$gamma\n";
	print "setmd max_num_cells\t100000";
	}

sub print_box {
# --------------
	print "# set up the simulation box\n";
	print "set box_l\t@_[0]\n";
	print "setmd box_l\t\$box_l \$box_l \$box_l\n";
	print "setmd periodic  1 1 1\n";
	print "puts \"set up simulation box with length \$box_l\\n\"\n\n";
	}

sub print_head { # The very first output
# --------------------------------------
	print "\#!/bin/sh\n";
	print "\# Tricking... the line after these comments are interpreted as standard shell script \\\\n";
	print "\texec \$ESPRESSO_SOURCE/Espresso \$0 \$*\n\n";

	print "set tcl_precision 6\n";
	print "\#source ./subs.tcl\n\n";

	print "# system name\n";
	print "set name  \"simulation\"\n\n";

	print "puts \" \"\n";
	print "puts \"=====================================================================\"\n";
	print "puts \"                           \$name.tcl                                 \"\n";
	print "puts \"=====================================================================\"\n";
	print "puts \" \"\n\n";

	print "\# output files\n";

	print "set traj_file [open \"\$name.vtf\" \"w\"]\n\n";
	}

sub generate_chain {
# -------------------
	my @chain = ("0 0 0");
	my $num_points = max(keys %{@_[0]});
	foreach my $bead ( sort { $a <=> $b } keys %{@_[0]} ) {
		next if ( $bead == 0 );
		my $bondLength = (${${@_[0]}{$bead}}{'radius'} + ${${@_[0]}{$bead-1}}{'radius'})/100;
		my $nextPoint = "0 0 $bondLength";
		push @chain,add(@chain[$#chain],$nextPoint);
		}
	return @chain;
	}
		
sub shift_chain {
# ----------------
		foreach my $i ( 0..$#{@_[0]} ) {
			my $new = add( ${@_[0]}[$i], @_[1] );
			${@_[0]}[$i] = $new;
		}
	}
	
sub subtract { # subtract the "first" @_[0] vector from the "second" @_[1]
# ------------------------------------------------------------------------
	my @a = split(" ",@_[0]);
	my @b = split(" ",@_[1]);
	my @c = ($a[0]-$b[0],$a[1]-$b[1],$a[2]-$b[2]);
	return join(" ",@c);
	}

sub flip { # flipping z-coord. from + to - with respect to z0-50 center
# ----------------------------------------------------------------------
	foreach my $point ( 0..$#{@_[0]} ) {
		my @point = split(" ",${@_[0]}[$point]);
		$point[2] -= 50;
		$point[2] = -1*$point[2];
		$point[2] += 50;
		${@_[0]}[$point] = join(" ",@point);
		}
	}
	
sub add { # add vectors
# ----------------------
	my @a = split(" ",@_[0]);
	my @b = split(" ",@_[1]);
	my @c = ($a[0]+$b[0],$a[1]+$b[1],$a[2]+$b[2]);
	return join(" ",@c);
	}
	
sub center_of_mass { # given an array of points "0 0 1" return the COM "0 0 5.5"
# ------------------------------------------------------------------------------
	my $points = @_[0];
	my $totalWeight = 0;
	for my $i ( 0..max(keys %{@_[1]}) ) {
		$totalWeight += ${${@_[1]}{$i}}{'weight'};
		}
	my @sum = (0,0,0);
	foreach my $p ( 0..$#{@_[0]} ) {
		my @point = split(" ",${@_[0]}[$p]);
		for my $i ( 0..$#point ) { $sum[$i] += $point[$i]*${${@_[1]}{$p}}{'weight'}; }
		}
	my @com = ($sum[0]/$totalWeight, $sum[1]/$totalWeight, $sum[2]/$totalWeight);
	return join(" ",@com);
	}
	
# Looks like this is for Pickersgill only. CHECK!!
sub count_lamin_binding {
# ------------------------
	my $count = 0;
	foreach my $site ( @{@_[0]} ) {
		if ( $site >= @_[1] ) {
			if ( $site <= @_[2] ) {
				$count++;
				}
			}
		}
	return $count;
	}

# NEW
sub tad_label {
# -------------
    my $label =  0;
    if (@_[0] eq "856" ) {$label = '11' ;}
    if (@_[0] eq "699" ) {$label = '11' ;}
    if (@_[0] eq "182" ) {$label = '12' ;}
    if (@_[0] eq "4" ) {$label = '12' ;}
    if (@_[0] eq "510" ) {$label = '13' ;}
    if (@_[0] eq "646" ) {$label = '13' ;}
    if (@_[0] eq "259" ) {$label = '14' ;}
    if (@_[0] eq "337" ) {$label = '14' ;}
    if (@_[0] eq "990" ) {$label = '15' ;}
    if (@_[0] eq "1116" ) {$label = '15' ;}
    if (@_[0] eq "449" ) {$label = '16' ;}
    if (@_[0] eq "458" ) {$label = '16' ;}
    if (@_[0] eq "206" ) {$label = '17' ;}
    if (@_[0] eq "219" ) {$label = '17' ;}
    if (@_[0] eq "409" ) {$label = '18' ;}
    if (@_[0] eq "258" ) {$label = '18' ;}
    if (@_[0] eq "1094" ) {$label = '19' ;}
    if (@_[0] eq "1150" ) {$label = '19' ;}
    if (@_[0] eq "564" ) {$label = '20' ;}
    if (@_[0] eq "619" ) {$label = '20' ;}
    if (@_[0] eq "495" ) {$label = '21' ;}
    if (@_[0] eq "525" ) {$label = '21' ;}
    if (@_[0] eq "21" ) {$label = '22' ;}
    if (@_[0] eq "134" ) {$label = '22' ;}
    if (@_[0] eq "232" ) {$label = '23' ;}
    if (@_[0] eq "234" ) {$label = '23' ;}
    if (@_[0] eq "154" ) {$label = '24' ;}
    if (@_[0] eq "208" ) {$label = '24' ;}
    if (@_[0] eq "875" ) {$label = '25' ;}
    if (@_[0] eq "933" ) {$label = '25' ;}
    if (@_[0] eq "570" ) {$label = '26' ;}
    if (@_[0] eq "578" ) {$label = '26' ;}
    if (@_[0] eq "669" ) {$label = '27' ;}
    if (@_[0] eq "749" ) {$label = '27' ;}
    if (@_[0] eq "154" ) {$label = '24' ;}
    if (@_[0] eq "203" ) {$label = '24' ;}
    if (@_[0] eq "564" ) {$label = '20' ;}
    if (@_[0] eq "640" ) {$label = '20' ;}
    if (@_[0] eq "570" ) {$label = '26' ;}
    if (@_[0] eq "587" ) {$label = '26' ;}
    if (@_[0] eq "990" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "12" ) {$label = '32' ;}
    if (@_[0] eq "29" ) {$label = '32' ;}
    if (@_[0] eq "70" ) {$label = '33' ;}
    if (@_[0] eq "113" ) {$label = '33' ;}
    if (@_[0] eq "1022" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "1034" ) {$label = '35' ;}
    if (@_[0] eq "1032" ) {$label = '35' ;}
    if (@_[0] eq "484" ) {$label = '36' ;}
    if (@_[0] eq "499" ) {$label = '36' ;}
    if (@_[0] eq "546" ) {$label = '37' ;}
    if (@_[0] eq "547" ) {$label = '37' ;}
    if (@_[0] eq "208" ) {$label = '24' ;}
    if (@_[0] eq "217" ) {$label = '24' ;}
    if (@_[0] eq "505" ) {$label = '39' ;}
    if (@_[0] eq "608" ) {$label = '39' ;}
    if (@_[0] eq "695" ) {$label = '40' ;}
    if (@_[0] eq "675" ) {$label = '40' ;}
    if (@_[0] eq "70" ) {$label = '33' ;}
    if (@_[0] eq "110" ) {$label = '33' ;}
    if (@_[0] eq "1030" ) {$label = '42' ;}
    if (@_[0] eq "1134" ) {$label = '42' ;}
    if (@_[0] eq "666" ) {$label = '43' ;}
    if (@_[0] eq "697" ) {$label = '43' ;}
    if (@_[0] eq "666" ) {$label = '43' ;}
    if (@_[0] eq "702" ) {$label = '43' ;}
    if (@_[0] eq "19" ) {$label = '45' ;}
    if (@_[0] eq "50" ) {$label = '45' ;}
    if (@_[0] eq "666" ) {$label = '43' ;}
    if (@_[0] eq "801" ) {$label = '43' ;}
    if (@_[0] eq "222" ) {$label = '47' ;}
    if (@_[0] eq "225" ) {$label = '47' ;}
    if (@_[0] eq "561" ) {$label = '48' ;}
    if (@_[0] eq "651" ) {$label = '48' ;}
    if (@_[0] eq "203" ) {$label = '24' ;}
    if (@_[0] eq "208" ) {$label = '24' ;}
    if (@_[0] eq "1065" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "695" ) {$label = '40' ;}
    if (@_[0] eq "660" ) {$label = '40' ;}
    if (@_[0] eq "392" ) {$label = '52' ;}
    if (@_[0] eq "438" ) {$label = '52' ;}
    if (@_[0] eq "990" ) {$label = '15' ;}
    if (@_[0] eq "1122" ) {$label = '15' ;}
    if (@_[0] eq "1022" ) {$label = '34' ;}
    if (@_[0] eq "1138" ) {$label = '34' ;}
    if (@_[0] eq "1022" ) {$label = '34' ;}
    if (@_[0] eq "1056" ) {$label = '34' ;}
    if (@_[0] eq "442" ) {$label = '56' ;}
    if (@_[0] eq "467" ) {$label = '56' ;}
    if (@_[0] eq "1062" ) {$label = '57' ;}
    if (@_[0] eq "1072" ) {$label = '57' ;}
    if (@_[0] eq "1113" ) {$label = '58' ;}
    if (@_[0] eq "1166" ) {$label = '58' ;}
    if (@_[0] eq "621" ) {$label = '59' ;}
    if (@_[0] eq "639" ) {$label = '59' ;}
    if (@_[0] eq "444" ) {$label = '60' ;}
    if (@_[0] eq "523" ) {$label = '60' ;}
    if (@_[0] eq "768" ) {$label = '61' ;}
    if (@_[0] eq "796" ) {$label = '61' ;}
    if (@_[0] eq "691" ) {$label = '43' ;}
    if (@_[0] eq "801" ) {$label = '43' ;}
    if (@_[0] eq "570" ) {$label = '26' ;}
    if (@_[0] eq "516" ) {$label = '26' ;}
    if (@_[0] eq "570" ) {$label = '26' ;}
    if (@_[0] eq "635" ) {$label = '26' ;}
    if (@_[0] eq "1123" ) {$label = '42' ;}
    if (@_[0] eq "1134" ) {$label = '42' ;}
    if (@_[0] eq "1116" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "1083" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '58' ;}
    if (@_[0] eq "1166" ) {$label = '58' ;}
    if (@_[0] eq "666" ) {$label = '61' ;}
    if (@_[0] eq "796" ) {$label = '61' ;}
    if (@_[0] eq "989" ) {$label = '70' ;}
    if (@_[0] eq "1120" ) {$label = '70' ;}
    if (@_[0] eq "661" ) {$label = '71' ;}
    if (@_[0] eq "692" ) {$label = '71' ;}
    if (@_[0] eq "690" ) {$label = '43' ;}
    if (@_[0] eq "702" ) {$label = '43' ;}
    if (@_[0] eq "454" ) {$label = '73' ;}
    if (@_[0] eq "468" ) {$label = '73' ;}
    if (@_[0] eq "208" ) {$label = '24' ;}
    if (@_[0] eq "221" ) {$label = '24' ;}
    if (@_[0] eq "491" ) {$label = '75' ;}
    if (@_[0] eq "593" ) {$label = '75' ;}
    if (@_[0] eq "154" ) {$label = '24' ;}
    if (@_[0] eq "210" ) {$label = '24' ;}
    if (@_[0] eq "578" ) {$label = '26' ;}
    if (@_[0] eq "587" ) {$label = '26' ;}
    if (@_[0] eq "260" ) {$label = '78' ;}
    if (@_[0] eq "295" ) {$label = '78' ;}
    if (@_[0] eq "157" ) {$label = '79' ;}
    if (@_[0] eq "171" ) {$label = '79' ;}
    if (@_[0] eq "796" ) {$label = '43' ;}
    if (@_[0] eq "801" ) {$label = '43' ;}
    if (@_[0] eq "566" ) {$label = '81' ;}
    if (@_[0] eq "641" ) {$label = '81' ;}
    if (@_[0] eq "55" ) {$label = '82' ;}
    if (@_[0] eq "127" ) {$label = '82' ;}
    if (@_[0] eq "1056" ) {$label = '34' ;}
    if (@_[0] eq "1138" ) {$label = '34' ;}
    if (@_[0] eq "1015" ) {$label = '84' ;}
    if (@_[0] eq "1118" ) {$label = '84' ;}
    if (@_[0] eq "694" ) {$label = '43' ;}
    if (@_[0] eq "697" ) {$label = '43' ;}
    if (@_[0] eq "727" ) {$label = '86' ;}
    if (@_[0] eq "886" ) {$label = '86' ;}
    if (@_[0] eq "562" ) {$label = '26' ;}
    if (@_[0] eq "570" ) {$label = '26' ;}
    if (@_[0] eq "1065" ) {$label = '50' ;}
    if (@_[0] eq "1055" ) {$label = '50' ;}
    if (@_[0] eq "570" ) {$label = '75' ;}
    if (@_[0] eq "593" ) {$label = '75' ;}
    if (@_[0] eq "998" ) {$label = '42' ;}
    if (@_[0] eq "1134" ) {$label = '42' ;}
    if (@_[0] eq "1072" ) {$label = '58' ;}
    if (@_[0] eq "1113" ) {$label = '58' ;}
    if (@_[0] eq "262" ) {$label = '92' ;}
    if (@_[0] eq "431" ) {$label = '92' ;}
    if (@_[0] eq "1058" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "933" ) {$label = '25' ;}
    if (@_[0] eq "940" ) {$label = '25' ;}
    if (@_[0] eq "586" ) {$label = '95' ;}
    if (@_[0] eq "583" ) {$label = '95' ;}
    if (@_[0] eq "1077" ) {$label = '96' ;}
    if (@_[0] eq "1114" ) {$label = '96' ;}
    if (@_[0] eq "1065" ) {$label = '84' ;}
    if (@_[0] eq "1015" ) {$label = '84' ;}
    if (@_[0] eq "1011" ) {$label = '98' ;}
    if (@_[0] eq "1018" ) {$label = '98' ;}
    if (@_[0] eq "594" ) {$label = '26' ;}
    if (@_[0] eq "635" ) {$label = '26' ;}
    if (@_[0] eq "559" ) {$label = '100' ;}
    if (@_[0] eq "652" ) {$label = '100' ;}
    if (@_[0] eq "449" ) {$label = '16' ;}
    if (@_[0] eq "457" ) {$label = '16' ;}
    if (@_[0] eq "219" ) {$label = '17' ;}
    if (@_[0] eq "223" ) {$label = '17' ;}
    if (@_[0] eq "594" ) {$label = '48' ;}
    if (@_[0] eq "651" ) {$label = '48' ;}
    if (@_[0] eq "1102" ) {$label = '57' ;}
    if (@_[0] eq "1072" ) {$label = '57' ;}
    if (@_[0] eq "562" ) {$label = '99' ;}
    if (@_[0] eq "594" ) {$label = '99' ;}
    if (@_[0] eq "529" ) {$label = '26' ;}
    if (@_[0] eq "635" ) {$label = '26' ;}
    if (@_[0] eq "1077" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '67' ;}
    if (@_[0] eq "182" ) {$label = '12' ;}
    if (@_[0] eq "186" ) {$label = '12' ;}
    if (@_[0] eq "1114" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '67' ;}
    if (@_[0] eq "1055" ) {$label = '50' ;}
    if (@_[0] eq "1057" ) {$label = '50' ;}
    if (@_[0] eq "990" ) {$label = '98' ;}
    if (@_[0] eq "1018" ) {$label = '98' ;}
    if (@_[0] eq "1114" ) {$label = '42' ;}
    if (@_[0] eq "1134" ) {$label = '42' ;}
    if (@_[0] eq "71" ) {$label = '82' ;}
    if (@_[0] eq "127" ) {$label = '82' ;}
    if (@_[0] eq "1015" ) {$label = '84' ;}
    if (@_[0] eq "1017" ) {$label = '84' ;}
    if (@_[0] eq "150" ) {$label = '115' ;}
    if (@_[0] eq "152" ) {$label = '115' ;}
    if (@_[0] eq "801" ) {$label = '43' ;}
    if (@_[0] eq "852" ) {$label = '43' ;}
    if (@_[0] eq "1082" ) {$label = '117' ;}
    if (@_[0] eq "1086" ) {$label = '117' ;}
    if (@_[0] eq "451" ) {$label = '118' ;}
    if (@_[0] eq "453" ) {$label = '118' ;}
    if (@_[0] eq "1097" ) {$label = '57' ;}
    if (@_[0] eq "1072" ) {$label = '57' ;}
    if (@_[0] eq "587" ) {$label = '26' ;}
    if (@_[0] eq "635" ) {$label = '26' ;}
    if (@_[0] eq "755" ) {$label = '121' ;}
    if (@_[0] eq "764" ) {$label = '121' ;}
    if (@_[0] eq "1017" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "1065" ) {$label = '34' ;}
    if (@_[0] eq "1022" ) {$label = '34' ;}
    if (@_[0] eq "1085" ) {$label = '67' ;}
    if (@_[0] eq "1083" ) {$label = '67' ;}
    if (@_[0] eq "635" ) {$label = '48' ;}
    if (@_[0] eq "651" ) {$label = '48' ;}
    if (@_[0] eq "1037" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "427" ) {$label = '52' ;}
    if (@_[0] eq "438" ) {$label = '52' ;}
    if (@_[0] eq "562" ) {$label = '87' ;}
    if (@_[0] eq "560" ) {$label = '87' ;}
    if (@_[0] eq "702" ) {$label = '61' ;}
    if (@_[0] eq "768" ) {$label = '61' ;}
    if (@_[0] eq "1120" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '67' ;}
    if (@_[0] eq "895" ) {$label = '131' ;}
    if (@_[0] eq "946" ) {$label = '131' ;}
    if (@_[0] eq "491" ) {$label = '75' ;}
    if (@_[0] eq "528" ) {$label = '75' ;}
    if (@_[0] eq "65" ) {$label = '133' ;}
    if (@_[0] eq "128" ) {$label = '133' ;}
    if (@_[0] eq "1015" ) {$label = '34' ;}
    if (@_[0] eq "1022" ) {$label = '34' ;}
    if (@_[0] eq "401" ) {$label = '135' ;}
    if (@_[0] eq "433" ) {$label = '135' ;}
    if (@_[0] eq "442" ) {$label = '118' ;}
    if (@_[0] eq "453" ) {$label = '118' ;}
    if (@_[0] eq "152" ) {$label = '79' ;}
    if (@_[0] eq "157" ) {$label = '79' ;}
    if (@_[0] eq "1160" ) {$label = '138' ;}
    if (@_[0] eq "1164" ) {$label = '138' ;}
    if (@_[0] eq "922" ) {$label = '139' ;}
    if (@_[0] eq "927" ) {$label = '139' ;}
    if (@_[0] eq "181" ) {$label = '140' ;}
    if (@_[0] eq "184" ) {$label = '140' ;}
    if (@_[0] eq "518" ) {$label = '81' ;}
    if (@_[0] eq "641" ) {$label = '81' ;}
    if (@_[0] eq "529" ) {$label = '26' ;}
    if (@_[0] eq "570" ) {$label = '26' ;}
    if (@_[0] eq "389" ) {$label = '143' ;}
    if (@_[0] eq "390" ) {$label = '143' ;}
    if (@_[0] eq "616" ) {$label = '20' ;}
    if (@_[0] eq "619" ) {$label = '20' ;}
    if (@_[0] eq "1030" ) {$label = '42' ;}
    if (@_[0] eq "1038" ) {$label = '42' ;}
    if (@_[0] eq "678" ) {$label = '146' ;}
    if (@_[0] eq "784" ) {$label = '146' ;}
    if (@_[0] eq "363" ) {$label = '92' ;}
    if (@_[0] eq "431" ) {$label = '92' ;}
    if (@_[0] eq "768" ) {$label = '43' ;}
    if (@_[0] eq "801" ) {$label = '43' ;}
    if (@_[0] eq "563" ) {$label = '26' ;}
    if (@_[0] eq "587" ) {$label = '26' ;}
    if (@_[0] eq "17" ) {$label = '45' ;}
    if (@_[0] eq "50" ) {$label = '45' ;}
    if (@_[0] eq "181" ) {$label = '140' ;}
    if (@_[0] eq "188" ) {$label = '140' ;}
    if (@_[0] eq "666" ) {$label = '43' ;}
    if (@_[0] eq "791" ) {$label = '43' ;}
    if (@_[0] eq "529" ) {$label = '81' ;}
    if (@_[0] eq "641" ) {$label = '81' ;}
    if (@_[0] eq "665" ) {$label = '71' ;}
    if (@_[0] eq "692" ) {$label = '71' ;}
    if (@_[0] eq "679" ) {$label = '62' ;}
    if (@_[0] eq "691" ) {$label = '62' ;}
    if (@_[0] eq "1114" ) {$label = '70' ;}
    if (@_[0] eq "1120" ) {$label = '70' ;}
    if (@_[0] eq "100" ) {$label = '82' ;}
    if (@_[0] eq "127" ) {$label = '82' ;}
    if (@_[0] eq "784" ) {$label = '86' ;}
    if (@_[0] eq "886" ) {$label = '86' ;}
    if (@_[0] eq "1038" ) {$label = '42' ;}
    if (@_[0] eq "1046" ) {$label = '42' ;}
    if (@_[0] eq "796" ) {$label = '43' ;}
    if (@_[0] eq "852" ) {$label = '43' ;}
    if (@_[0] eq "260" ) {$label = '92' ;}
    if (@_[0] eq "431" ) {$label = '92' ;}
    if (@_[0] eq "562" ) {$label = '13' ;}
    if (@_[0] eq "646" ) {$label = '13' ;}
    if (@_[0] eq "76" ) {$label = '82' ;}
    if (@_[0] eq "127" ) {$label = '82' ;}
    if (@_[0] eq "1095" ) {$label = '164' ;}
    if (@_[0] eq "1148" ) {$label = '164' ;}
    if (@_[0] eq "50" ) {$label = '45' ;}
    if (@_[0] eq "8" ) {$label = '45' ;}
    if (@_[0] eq "1155" ) {$label = '138' ;}
    if (@_[0] eq "1160" ) {$label = '138' ;}
    if (@_[0] eq "1090" ) {$label = '57' ;}
    if (@_[0] eq "1072" ) {$label = '57' ;}
    if (@_[0] eq "184" ) {$label = '12' ;}
    if (@_[0] eq "186" ) {$label = '12' ;}
    if (@_[0] eq "1050" ) {$label = '57' ;}
    if (@_[0] eq "1072" ) {$label = '57' ;}
    if (@_[0] eq "875" ) {$label = '25' ;}
    if (@_[0] eq "961" ) {$label = '25' ;}
    if (@_[0] eq "455" ) {$label = '73' ;}
    if (@_[0] eq "468" ) {$label = '73' ;}
    if (@_[0] eq "33" ) {$label = '82' ;}
    if (@_[0] eq "127" ) {$label = '82' ;}
    if (@_[0] eq "852" ) {$label = '43' ;}
    if (@_[0] eq "853" ) {$label = '43' ;}
    if (@_[0] eq "360" ) {$label = '135' ;}
    if (@_[0] eq "433" ) {$label = '135' ;}
    if (@_[0] eq "323" ) {$label = '175' ;}
    if (@_[0] eq "242" ) {$label = '175' ;}
    if (@_[0] eq "81" ) {$label = '82' ;}
    if (@_[0] eq "127" ) {$label = '82' ;}
    if (@_[0] eq "100" ) {$label = '157' ;}
    if (@_[0] eq "102" ) {$label = '157' ;}
    if (@_[0] eq "1030" ) {$label = '42' ;}
    if (@_[0] eq "1028" ) {$label = '42' ;}
    if (@_[0] eq "1049" ) {$label = '57' ;}
    if (@_[0] eq "1072" ) {$label = '57' ;}
    if (@_[0] eq "666" ) {$label = '61' ;}
    if (@_[0] eq "768" ) {$label = '61' ;}
    if (@_[0] eq "887" ) {$label = '181' ;}
    if (@_[0] eq "947" ) {$label = '181' ;}
    if (@_[0] eq "721" ) {$label = '182' ;}
    if (@_[0] eq "786" ) {$label = '182' ;}
    if (@_[0] eq "1067" ) {$label = '183' ;}
    if (@_[0] eq "1073" ) {$label = '183' ;}
    if (@_[0] eq "157" ) {$label = '79' ;}
    if (@_[0] eq "201" ) {$label = '79' ;}
    if (@_[0] eq "1048" ) {$label = '42' ;}
    if (@_[0] eq "1134" ) {$label = '42' ;}
    if (@_[0] eq "749" ) {$label = '27' ;}
    if (@_[0] eq "751" ) {$label = '27' ;}
    if (@_[0] eq "528" ) {$label = '75' ;}
    if (@_[0] eq "592" ) {$label = '75' ;}
    if (@_[0] eq "739" ) {$label = '188' ;}
    if (@_[0] eq "742" ) {$label = '188' ;}
    if (@_[0] eq "663" ) {$label = '146' ;}
    if (@_[0] eq "678" ) {$label = '146' ;}
    if (@_[0] eq "493" ) {$label = '190' ;}
    if (@_[0] eq "517" ) {$label = '190' ;}
    if (@_[0] eq "730" ) {$label = '191' ;}
    if (@_[0] eq "732" ) {$label = '191' ;}
    if (@_[0] eq "682" ) {$label = '192' ;}
    if (@_[0] eq "696" ) {$label = '192' ;}
    if (@_[0] eq "255" ) {$label = '193' ;}
    if (@_[0] eq "340" ) {$label = '193' ;}
    if (@_[0] eq "754" ) {$label = '194' ;}
    if (@_[0] eq "752" ) {$label = '194' ;}
    if (@_[0] eq "2" ) {$label = '157' ;}
    if (@_[0] eq "100" ) {$label = '157' ;}
    if (@_[0] eq "518" ) {$label = '87' ;}
    if (@_[0] eq "560" ) {$label = '87' ;}
    if (@_[0] eq "562" ) {$label = '141' ;}
    if (@_[0] eq "518" ) {$label = '141' ;}
    if (@_[0] eq "562" ) {$label = '87' ;}
    if (@_[0] eq "632" ) {$label = '87' ;}
    if (@_[0] eq "366" ) {$label = '92' ;}
    if (@_[0] eq "431" ) {$label = '92' ;}
    if (@_[0] eq "468" ) {$label = '73' ;}
    if (@_[0] eq "452" ) {$label = '73' ;}
    if (@_[0] eq "150" ) {$label = '115' ;}
    if (@_[0] eq "147" ) {$label = '115' ;}
    if (@_[0] eq "24" ) {$label = '202' ;}
    if (@_[0] eq "9" ) {$label = '202' ;}
    if (@_[0] eq "54" ) {$label = '203' ;}
    if (@_[0] eq "129" ) {$label = '203' ;}
    if (@_[0] eq "8" ) {$label = '45' ;}
    if (@_[0] eq "43" ) {$label = '45' ;}
    if (@_[0] eq "1031" ) {$label = '205' ;}
    if (@_[0] eq "1047" ) {$label = '205' ;}
    if (@_[0] eq "491" ) {$label = '75' ;}
    if (@_[0] eq "504" ) {$label = '75' ;}
    if (@_[0] eq "523" ) {$label = '100' ;}
    if (@_[0] eq "559" ) {$label = '100' ;}
    if (@_[0] eq "76" ) {$label = '163' ;}
    if (@_[0] eq "141" ) {$label = '163' ;}
    if (@_[0] eq "560" ) {$label = '26' ;}
    if (@_[0] eq "635" ) {$label = '26' ;}
    if (@_[0] eq "1139" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '67' ;}
    if (@_[0] eq "246" ) {$label = '211' ;}
    if (@_[0] eq "248" ) {$label = '211' ;}
    if (@_[0] eq "1113" ) {$label = '164' ;}
    if (@_[0] eq "1148" ) {$label = '164' ;}
    if (@_[0] eq "54" ) {$label = '203' ;}
    if (@_[0] eq "107" ) {$label = '203' ;}
    if (@_[0] eq "891" ) {$label = '181' ;}
    if (@_[0] eq "887" ) {$label = '181' ;}
    if (@_[0] eq "1121" ) {$label = '164' ;}
    if (@_[0] eq "1148" ) {$label = '164' ;}
    if (@_[0] eq "685" ) {$label = '40' ;}
    if (@_[0] eq "660" ) {$label = '40' ;}
    if (@_[0] eq "664" ) {$label = '40' ;}
    if (@_[0] eq "695" ) {$label = '40' ;}
    if (@_[0] eq "41" ) {$label = '12' ;}
    if (@_[0] eq "4" ) {$label = '12' ;}
    if (@_[0] eq "349" ) {$label = '92' ;}
    if (@_[0] eq "431" ) {$label = '92' ;}
    if (@_[0] eq "399" ) {$label = '92' ;}
    if (@_[0] eq "431" ) {$label = '92' ;}
    if (@_[0] eq "1110" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '67' ;}
    if (@_[0] eq "261" ) {$label = '222' ;}
    if (@_[0] eq "292" ) {$label = '222' ;}
    if (@_[0] eq "3" ) {$label = '223' ;}
    if (@_[0] eq "53" ) {$label = '223' ;}
    if (@_[0] eq "510" ) {$label = '75' ;}
    if (@_[0] eq "491" ) {$label = '75' ;}
    if (@_[0] eq "363" ) {$label = '147' ;}
    if (@_[0] eq "432" ) {$label = '147' ;}
    if (@_[0] eq "1017" ) {$label = '34' ;}
    if (@_[0] eq "1022" ) {$label = '34' ;}
    if (@_[0] eq "922" ) {$label = '139' ;}
    if (@_[0] eq "938" ) {$label = '139' ;}
    if (@_[0] eq "660" ) {$label = '40' ;}
    if (@_[0] eq "677" ) {$label = '40' ;}
    if (@_[0] eq "998" ) {$label = '42' ;}
    if (@_[0] eq "1038" ) {$label = '42' ;}
    if (@_[0] eq "1072" ) {$label = '57' ;}
    if (@_[0] eq "1074" ) {$label = '57' ;}
    if (@_[0] eq "223" ) {$label = '47' ;}
    if (@_[0] eq "225" ) {$label = '47' ;}
    if (@_[0] eq "201" ) {$label = '79' ;}
    if (@_[0] eq "196" ) {$label = '79' ;}
    if (@_[0] eq "752" ) {$label = '194' ;}
    if (@_[0] eq "785" ) {$label = '194' ;}
    if (@_[0] eq "366" ) {$label = '199' ;}
    if (@_[0] eq "370" ) {$label = '199' ;}
    if (@_[0] eq "988" ) {$label = '90' ;}
    if (@_[0] eq "998" ) {$label = '90' ;}
    if (@_[0] eq "1120" ) {$label = '42' ;}
    if (@_[0] eq "1134" ) {$label = '42' ;}
    if (@_[0] eq "1072" ) {$label = '96' ;}
    if (@_[0] eq "1114" ) {$label = '96' ;}
    if (@_[0] eq "875" ) {$label = '25' ;}
    if (@_[0] eq "940" ) {$label = '25' ;}
    if (@_[0] eq "24" ) {$label = '45' ;}
    if (@_[0] eq "8" ) {$label = '45' ;}
    if (@_[0] eq "1027" ) {$label = '205' ;}
    if (@_[0] eq "1047" ) {$label = '205' ;}
    if (@_[0] eq "654" ) {$label = '241' ;}
    if (@_[0] eq "657" ) {$label = '241' ;}
    if (@_[0] eq "632" ) {$label = '48' ;}
    if (@_[0] eq "651" ) {$label = '48' ;}
    if (@_[0] eq "406" ) {$label = '18' ;}
    if (@_[0] eq "409" ) {$label = '18' ;}
    if (@_[0] eq "923" ) {$label = '244' ;}
    if (@_[0] eq "928" ) {$label = '244' ;}
    if (@_[0] eq "1036" ) {$label = '67' ;}
    if (@_[0] eq "1152" ) {$label = '67' ;}
    if (@_[0] eq "998" ) {$label = '90' ;}
    if (@_[0] eq "1005" ) {$label = '90' ;}
    if (@_[0] eq "727" ) {$label = '194' ;}
    if (@_[0] eq "785" ) {$label = '194' ;}
    if (@_[0] eq "66" ) {$label = '203' ;}
    if (@_[0] eq "107" ) {$label = '203' ;}
    if (@_[0] eq "688" ) {$label = '249' ;}
    if (@_[0] eq "684" ) {$label = '249' ;}
    if (@_[0] eq "922" ) {$label = '139' ;}
    if (@_[0] eq "934" ) {$label = '139' ;}
    if (@_[0] eq "925" ) {$label = '251' ;}
    if (@_[0] eq "929" ) {$label = '251' ;}
    if (@_[0] eq "1153" ) {$label = '138' ;}
    if (@_[0] eq "1160" ) {$label = '138' ;}
    if (@_[0] eq "184" ) {$label = '140' ;}
    if (@_[0] eq "188" ) {$label = '140' ;}
    if (@_[0] eq "1160" ) {$label = '58' ;}
    if (@_[0] eq "1166" ) {$label = '58' ;}
    if (@_[0] eq "166" ) {$label = '255' ;}
    if (@_[0] eq "170" ) {$label = '255' ;}
    if (@_[0] eq "493" ) {$label = '75' ;}
    if (@_[0] eq "504" ) {$label = '75' ;}
    if (@_[0] eq "76" ) {$label = '157' ;}
    if (@_[0] eq "100" ) {$label = '157' ;}
    if (@_[0] eq "234" ) {$label = '23' ;}
    if (@_[0] eq "234" ) {$label = '23' ;}
    if (@_[0] eq "678" ) {$label = '71' ;}
    if (@_[0] eq "692" ) {$label = '71' ;}
    if (@_[0] eq "927" ) {$label = '139' ;}
    if (@_[0] eq "934" ) {$label = '139' ;}
    if (@_[0] eq "1044" ) {$label = '42' ;}
    if (@_[0] eq "1046" ) {$label = '42' ;}
    if (@_[0] eq "1113" ) {$label = '65' ;}
    if (@_[0] eq "1123" ) {$label = '65' ;}
    if (@_[0] eq "640" ) {$label = '241' ;}
    if (@_[0] eq "657" ) {$label = '241' ;}
    if (@_[0] eq "746" ) {$label = '188' ;}
    if (@_[0] eq "739" ) {$label = '188' ;}
    if (@_[0] eq "504" ) {$label = '75' ;}
    if (@_[0] eq "522" ) {$label = '75' ;}
    if (@_[0] eq "1035" ) {$label = '266' ;}
    if (@_[0] eq "999" ) {$label = '266' ;}
    if (@_[0] eq "1131" ) {$label = '15' ;}
    if (@_[0] eq "1133" ) {$label = '15' ;}
    if (@_[0] eq "458" ) {$label = '16' ;}
    if (@_[0] eq "457" ) {$label = '16' ;}
    if (@_[0] eq "510" ) {$label = '13' ;}
    if (@_[0] eq "557" ) {$label = '13' ;}
    if (@_[0] eq "554" ) {$label = '13' ;}
    if (@_[0] eq "557" ) {$label = '13' ;}
    if (@_[0] eq "389" ) {$label = '143' ;}
    if (@_[0] eq "394" ) {$label = '143' ;}
    if (@_[0] eq "14" ) {$label = '172' ;}
    if (@_[0] eq "33" ) {$label = '172' ;}
    if (@_[0] eq "739" ) {$label = '188' ;}
    if (@_[0] eq "744" ) {$label = '188' ;}
    if (@_[0] eq "24" ) {$label = '202' ;}
    if (@_[0] eq "27" ) {$label = '202' ;}
    if (@_[0] eq "587" ) {$label = '26' ;}
    if (@_[0] eq "591" ) {$label = '26' ;}
    if (@_[0] eq "1024" ) {$label = '261' ;}
    if (@_[0] eq "1044" ) {$label = '261' ;}
    if (@_[0] eq "424" ) {$label = '277' ;}
    if (@_[0] eq "421" ) {$label = '277' ;}
    if (@_[0] eq "197" ) {$label = '278' ;}
    if (@_[0] eq "198" ) {$label = '278' ;}
    return $label;
    }


#my $key = 1;
###if ( grep { $_ == $label } in [237,245] { return \%{ $key++ => .647} }
###if ( grep { $_ == $label } in [800,666,547] { return \%{ $key++ => .5} }
###if ( grep { $_ == $label } in [800,666,547] { return \%{ $key++ => .5} }
